//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#include "glBoy.hh"
#include "Core.hh"
#include "ALU.hh"
#include "DAA.hh"
#include "Log.hh"
#include "util.hh"

#include <cassert>
#include <limits>
#include <sstream>

using namespace glBoy;

Core::Core() :
	m_iff1(0),
	m_iff2(0),
	m_halted(false),
	m_clock(0),
	m_nextPeriodicCallback(std::numeric_limits<cycle_t>::max())
{
	m_reg.reset();
	ALU::load();
}

void
Core::raiseInterrupt(int interrupt)
{
	if (m_iff1 != 0) // Are interrupts enabled ?
	{
		// TODO this interrupt handling is GB80 specific
		bool enabled(false);

		uint8_t flags(m_mem.read8(htoz(0xFFFF)));
		uint8_t outFlags(m_mem.read8(htoz(0xFF0F)));

		switch (interrupt)
		{
		case 0x40: enabled = flags & 0x1; outFlags |= 0x1; break;
		case 0x48: enabled = flags & 0x2; outFlags |= 0x2; break;
		case 0x50: enabled = flags & 0x4; outFlags |= 0x4; break;
		case 0x58: enabled = flags & 0x8; outFlags |= 0x8; break;
		case 0x60: enabled = flags & 0x10; outFlags += 0x10; break;
		}
		if (enabled)
		{
			m_interruptQueue.push_back(interrupt);
		}
	}
}

void
Core::run()
{
#include "InstructionSet_Tables.cc"

	// Instruction decoding occurs in an odd order for the
	// double-prefix instructions. PREFIX PREFIX OPERAND OPCODE
	// This operand value is set in the dispatch method of the double-opcode
	// handler
	// TODO z80 only. Unused for gb80 uint8_t doublePrefixOperand(0);

	Log& log(Log::Instance());

	while (true)
	{
		cycle_t clock = 0;

		uint16_t prefix(0); // Used for debug logging only.

		// Check for interrupts.
		if (!m_interruptQueue.empty())
		{
			int interrupt = m_interruptQueue.front();
			m_interruptQueue.pop_front();

			m_iff1 = 0;
			m_iff2 = 0;

			// TODO interrupt modes!
			m_reg.SP.dec(2);
			m_mem.write16(m_reg.SP, m_reg.PC);

			m_reg.PC = htoz(interrupt);

			m_halted = false;
		}
		else if (!m_halted)
		{
			uint16_t opcodePC = m_reg.PC.host(); // For debugging
			uint8_t opcode = op8();

			goto *(OpcodeTable[opcode]);

#include "InstructionSet_Opcodes.cc"

BAD_OPCODE:
			log(Log::DBG_ERROR) <<
				"Unknown Opcode: " << glBoy::hex(opcode) << " (PC=" << opcodePC << ")" <<
				std::endl;
			assert(false);

END_INSTRUCTIONS: ;
		}
		else
		{
			clock = 4;
		}

		m_clock += clock;

		if (m_halted && m_interruptQueue.empty())
		{
			m_clock = std::max(m_clock, m_nextPeriodicCallback);
		}

		if (m_clock >= m_nextPeriodicCallback)
		{
			m_nextPeriodicCallback = std::numeric_limits<cycle_t>::max();
			for (size_t i = 0; i < m_periodicCallbacks.size(); ++i)
			{
				if (m_clock >= m_periodicCallbacks[i].nextPeriod)
				{
					m_periodicCallbacks[i].callback(i);

					while (m_periodicCallbacks[i].nextPeriod <= m_clock)
					{
						m_periodicCallbacks[i].nextPeriod +=
							m_periodicCallbacks[i].clocks;
					}
				}

				if (m_periodicCallbacks[i].nextPeriod < m_nextPeriodicCallback)
				{
					m_nextPeriodicCallback = m_periodicCallbacks[i].nextPeriod;
				}
			}
		}
	}
}

Core::CallbackId
Core::registerPeriodic(
	int clockPeriod, int offset, const PeriodicCallbackFn& callback
	)
{
	assert(m_clock == 0);
	PeriodicCallback periodic = { clockPeriod, offset, callback };

	if (periodic.nextPeriod < m_nextPeriodicCallback)
	{
		m_nextPeriodicCallback = periodic.nextPeriod;
	}

	m_periodicCallbacks.push_back(periodic);
	return m_periodicCallbacks.size() - 1;
}

uint8_t Core::op8()
{
	uint8_t result(m_mem.read8(m_reg.PC));
	m_reg.PC.inc();
	return result;
}

DWORD Core::op16()
{
	DWORD operand(m_mem.read16(m_reg.PC));
	m_reg.PC.inc(2);
	return operand;
}

