//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#include "glBoy.hh"
#include "GameboyCart.hh"
#include "RAM.hh"
#include "ROM.hh"

#include <cassert>

using namespace glBoy;
using namespace std;

namespace
{
	class MBC1 : public MemoryMap::Memory
	{
	public:
		MBC1(GameboyCart* parent, uint8_t* begin, uint8_t* end) :
			m_parent(parent),
			m_rom(begin, end),
			m_mode(ROM_SELECT),
			m_romLow(1),
			m_romHigh(0),
			m_romOffset(0)
		{
		}

		virtual uint8_t read8(uint16_t address)
		{
			if (address < 0x4000)
			{
				return m_rom[address];
			}
			else if ((m_romOffset + address) < m_rom.size())
			{
				return m_rom[m_romOffset + address];
			}
			else
			{
std::cerr << address << "  " << m_rom.size() << " " << m_romOffset << " " << (int) m_romLow << " " << (int) m_romHigh << std::endl;
assert(false);
				return 0;
			}
		}

		virtual void write8(uint16_t address, uint8_t value)
		{
			if (address < 0x1FFF)
			{
				// TODO enable/disable ram
			}
			else if (address < 0x4000)
			{
				m_romLow = (value == 0) ? 1 : value;

				m_romOffset =
					((((m_romHigh & 0x3) << 4) | (m_romLow & 0x1F)) - 1) * 0x4000;
assert(m_romOffset < m_rom.size());
			}
			else if (address < 0x6000)
			{
				if (m_mode == RAM_SELECT)
				{
					m_parent->setRamBank(value & 0x3);
				}
				else
				{
					m_romHigh = value;

					m_romOffset =
						((((m_romHigh & 0x3) << 4) | (m_romLow & 0x1F)) - 1) * 0x4000;
assert(m_romOffset < m_rom.size());
				}
			}
			else
			{
				if (value & 1)
				{
					m_mode = RAM_SELECT;
				}
				else
				{
					m_mode = ROM_SELECT;
					m_parent->setRamBank(0);
				}
			}
		}

	private:
		GameboyCart* m_parent;

		std::vector<uint16_t> m_rom;

		enum Mode { ROM_SELECT, RAM_SELECT };
		Mode m_mode;

		uint8_t m_romLow;
		uint8_t m_romHigh;

		uint32_t m_romOffset;
	};
}

GameboyCart::GameboyCart(
	Core& core, uint8_t* rom, size_t size
	) :
	m_core(core)
{
	if (size < 16384)
	{
		std::cerr << "Invalid ROM (too small)" << std::endl;
		abort();
	}

	bool error(false);

	string mbc;
	switch (rom[0x147])
	{
	case 0:
	{
		mbc = "None";

		shared_ptr<MemoryMap::Memory> mem(
			new MBC1(this, rom, rom + size)
			);
/*		shared_ptr<MemoryMap::Memory> mem(
			new ROM(rom, size)
			);*/
		m_core.getMemoryMap().map(0, 0x8000, mem);
		m_romBanks.push_back(mem);
	}; break;

	case 1:
	case 2:
	case 3:
	{
		mbc = "MBC1";
		shared_ptr<MemoryMap::Memory> mem(
			new MBC1(this, rom, rom + size)
			);
		m_core.getMemoryMap().map(0, 0x8000, mem);
		m_romBanks.push_back(mem);
	};
	break;
	default:
		mbc = "Unsupported";
		error = true;
	};

	int cartRam;
	switch (rom[0x149])
	{
	case 0:
	{
		// Mis-reported on some carts.
		cartRam = 0;
		shared_ptr<MemoryMap::Memory> mem(new RAM(8192));
		m_core.getMemoryMap().map(0xA000, 8192, mem);
		m_ramBanks.push_back(mem);
	}; break;
	case 1:
	{
		cartRam = 2048;;

		// Fill out the entire 8k, just in case.
		shared_ptr<MemoryMap::Memory> mem(new RAM(8192));
		m_core.getMemoryMap().map(0xA000, cartRam, mem);
		m_ramBanks.push_back(mem);
	}; break;

	case 2:
	{
		cartRam = 8192;
		shared_ptr<MemoryMap::Memory> mem(new RAM(cartRam));
		m_core.getMemoryMap().map(0xA000, cartRam, mem);
		m_ramBanks.push_back(mem);
	}; break;

	case 3:
	{
		cartRam = 32768;
		for (int i = 0; i < 4; ++i)
		{
			shared_ptr<MemoryMap::Memory> mem(new RAM(8192));
			m_ramBanks.push_back(mem);
		}
		m_core.getMemoryMap().map(0xA000, 8192, m_ramBanks.front());
	}; break;

	default:
		error = true;
		cartRam = 0;
		cerr << "Invalid RAM size" << endl;
	}

	string title(rom + 0x134, rom + 0x134 + 11);
	cout << "Loading Cart" << endl;
	cout << "\tTitle:\t\t" << title << endl;
	cout << "\tRegion:\t\t" << (rom[0x14A] ? "Worldwide" : "Japan") << endl;
	cout << "\tVersion:\t" << int(rom[0x14D]) << endl;
	cout << "\tMBC:\t\t" << mbc << endl;
	cout << "\tROM size:\t" << (32 << rom[0x148]) << "KB" << endl;
	cout << "\tExt RAM size:\t" << (cartRam / 1024) << "KB" << endl;

	if (error)
	{
		abort();
	}
}

void
GameboyCart::setRamBank(size_t bank)
{
	if (bank < m_ramBanks.size())
	{
		m_core.getMemoryMap().remove(0xA000);
		m_core.getMemoryMap().map(0xA000, 8192, m_ramBanks[bank]);
	}
}
