//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#ifndef GLBOY_GAMEBOYGRAPHICS_HH
#define GLBOY_GAMEBOYGRAPHICS_HH

#include "glBoy.hh"
#include "Core.hh"

#include "MemoryMap.hh"

#include "hq4x.hh"

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include <vector>

namespace glBoy
{

class GameboyGraphics
{
public:
	enum PixelScaler { Scaler_None, Scaler_Scale2x, Scaler_hq4x };
	enum ResampleMethod { Resample_Nearest, Resample_Bilinear };

	GameboyGraphics(
		Core& core,
		SDL_Surface* surface,
		PixelScaler scaler,
		ResampleMethod resampler
		);
	~GameboyGraphics();

	void startScanLine();
	void renderScanLine();

	void setFrameSkip() { m_frameSkip = true; }
	void clearFrameSkip() { m_frameSkip = false; }

private:
	void periodicCallback(Core::CallbackId);

	void backgroundChanged() { m_backgroundRedraw = true; }
	void oamChanged() {}

	struct Pixel // Must be POD - used with memset.
	{
		int16_t tileX;  // Max for background/window
		uint8_t palette;
		uint8_t colourIndex; // Index into pallete.
	};

	struct FrameBuffer
	{
		int width;
		int height;
		Pixel* frame;
	};

	// RENDER_BLIT = overwrite background.
	// RENDER_SPRITE = only overwrite background colour 0 (ie. send
	// behind background)
	enum RenderMode{ RENDER_BLIT, RENDER_SPRITE };

	void renderScreen();
	void displayStats();

	void renderTile(
		uint8_t tile,
		int x,
		int y,
		const FrameBuffer& frameBuffer
		);

	void renderTile(
		uint16_t tileAddr,
		int x,
		int y,
		const FrameBuffer& frameBuffer,
		RenderMode mode,
		uint8_t pallete,
		uint8_t sprite = 0, // 0x80 -> 0xA8, 0 for background/window
		bool horizFlip = false,
		bool vertFlip = false,
		int tileHeight = 8 // 8 or 16
		);

	struct Sprite
	{
		int x;
		int y;
		uint8_t tile;
		uint8_t flags;
	};
	Sprite getSprite(int spriteNum) const;

	void blitGreyscale(uint8_t* dest);
	void updateOutputDevice();

	Core& m_core;

	PixelScaler m_scaler;
	ResampleMethod m_resampler;

	Hq4x m_hq4x;

	// Graphics system operates with 8x8 pixel tiles. There is space for 384
	// tiles. 256 tiles map be used in a map. Map 1 uses tile numbers 0 to 255,
	// other uses -128 to 127.
	// There are 2 maps held in memory, but only one displayed at a time.
	// 0x8000 - 0x87FF: Tile Set #1: 0 - 127.
	// 0x8800 - 0x8FFF: Tile Set #1: 128-255.  Tile Set #0: -1 to -128
	// 0x9000 - 0x97FF: Tile Set #0: tiles 0-127
	// 0x9800 - 0x9BFF: Tile map #0
	// 0x9C00 - 0x9FFF: Tile map #1
	uint8_t* m_vram;

	// Object Attribute Memory. Up to 40 sprites, which reference first Tile Set
	uint8_t* m_oam; // 0xFE00 - 0xFE9F

	// Memory-mapped IO Registers
	class GraphicsRegisters;
	std::shared_ptr<GraphicsRegisters> m_reg;

	// Raster display
	bool m_backgroundRedraw;
	Pixel m_background[256][256];
	Pixel m_screen[144][160];

	// Output device
	SDL_Surface* m_surface;

	// Shader program
	GLint m_shaderProgram;

	Core::CallbackId m_mode0;
	Core::CallbackId m_mode2;
	Core::CallbackId m_mode3;

	int m_lineScrollX[154];
	int m_lineScrollY[154];


	bool m_frameSkip;
	bool m_lastSkipped;

	timeval m_stat_fpsTimer;
	int m_stat_frames;
	int m_stat_skipped;
	int m_stat_redraw;
};

} // namespace glBoy

#endif

