//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#include "glBoy.hh"
#include "GameboySound.hh"

#include <cassert>
#include <cstdlib>
#include <cmath>

using namespace glBoy;

enum { SAMPLE_RATE = 48000 };

extern "C"
{
	void mixaudio(void *userData, Uint8 *stream, int len)
	{
		GameboySound* ourObject(reinterpret_cast<GameboySound*>(userData));
		ourObject->mixSounds(stream, len);
	}
}

GameboySound::GameboySound(Core& core)
{
	extern void mixaudio(void *unused, Uint8 *stream, int len);
	SDL_AudioSpec fmt;

	fmt.freq = SAMPLE_RATE;
	fmt.format = AUDIO_S16SYS;
	fmt.channels = 1;
	fmt.samples = 512;
	fmt.callback = mixaudio;
	fmt.userdata = this;

	if ( SDL_OpenAudio(&fmt, NULL) < 0 )
	{
		assert(false);
	}
	SDL_PauseAudio(0);

	using namespace std::placeholders;
	core.registerPeriodic(
		32768, // 128 Hz
		0,
		std::bind(&glBoy::GameboySound::periodicUpdate, this, _1)
		);
}

uint8_t
GameboySound::read8(uint16_t address)
{
	uint8_t result(0);

	switch (address)
	{
	case (0x0):
		result = m_channel1.sweep;
		break;
	case (0x1):
		result = m_channel1.duty;
		break;
	case (0x2):
		result = m_channel1.volume;
		break;
	case (0x3):
		result = m_channel1.freqLow;
		break;
	case (0x4):
		result = m_channel1.freqHigh;
		break;

	case (0x6):
		result = m_channel2.duty;
		break;
	case (0x7):
		result = m_channel2.volume;
		break;
	case (0x8):
		result = m_channel2.freqLow;
		break;
	case (0x9):
		result = m_channel2.freqHigh;
		break;

	case (0xA):
		result = m_channel3.enabled;
		break;
	case (0xB):
		result = m_channel3.length;
		break;
	case (0xC):
		result = m_channel3.volume;
		break;
	case (0xD):
		result = m_channel3.freqLow;
		break;
	case (0xE):
		result = m_channel3.freqHigh;
		break;

	case (0x10):
		result = m_channel4.length;
		break;
	case (0x11):
		result = m_channel4.volume;
		break;
	case (0x12):
		result = m_channel4.poly;
		break;
	case (0x13):
		result = m_channel4.counter;
		break;
	// TODO REMOVE THIS AND IMPLEMENT EVERYTHING.
	};

	if (address >= 0x20 && address < 0x30)
	{
		result = m_channel3.waveTable[address - 0x20];
	}
	return result;
}

void
GameboySound::write8(uint16_t address, uint8_t value)
{
	switch (address)
	{
	case (0x0):
		m_channel1.sweep = value;
		m_channel1.sweepDelay = 0;
		m_channel1.sweepCount = 0;
		break;
	case (0x1):
		m_channel1.duty = value;
		break;
	case (0x2):
		m_channel1.volume = value;
		break;
	case (0x3):
		m_channel1.freqLow = value;
		break;
	case (0x4):
		m_channel1.freqHigh = value;

		if (value & 0x80)
		{
			m_channel1.reset();
		}
		break;

	case (0x6):
		m_channel2.duty = value;
		break;
	case (0x7):
		m_channel2.volume = value;
		break;
	case (0x8):
		m_channel2.freqLow = value;
		break;
	case (0x9):
		m_channel2.freqHigh = value;
		if (value & 0x80)
		{
			m_channel2.reset();
		}
		break;

	case (0xA):
		m_channel3.enabled = value;
		break;
	case (0xB):
		m_channel3.length= value;
		break;
	case (0xC):
		m_channel3.volume = value;
		break;
	case (0xD):
		m_channel3.freqLow = value;
		break;
	case (0xE):
		m_channel3.freqHigh = value;
		if (value & 0x80)
		{
			m_channel3.reset();
		}
		break;


	case (0x10):
		m_channel4.length= value;
		break;
	case (0x11):
		m_channel4.volume = value;
		break;
	case (0x12):
		m_channel4.poly= value;
		break;
	case (0x13):
		m_channel4.counter = value;
		if (value & 0x80)
		{
			m_channel4.reset();
		}
		break;

	// TODO REMOVE THIS AND IMPLEMENT EVERYTHING.
	};

	if (address >= 0x20 && address < 0x30)
	{
		m_channel3.waveTable[address - 0x20] = value;
	}
}

void GameboySound::periodicUpdate(Core::CallbackId)
{
	prepChannel(m_channel1);
	prepChannel(m_channel2);
	prepChannel3();
	prepChannel4();
}

void
GameboySound::mixSounds(Uint8 *stream, int len)
{
	for (int i = 0; i < len / 2; ++i)
	{
		int32_t val(0);
		val += sample(m_channel1);
		val += sample(m_channel2);
		val += sampleChannel3();
		val += sampleChannel4();
		val >>= 2;
		if (val > 32767) val = 32767;
		if (val < -32768) val = -32768;

//TODO ENDIAN FIAL HERE
		stream[i * 2] = val & 0xff;
		stream[i * 2 + 1] = val >> 8;
	}

}

int16_t
GameboySound::sample(SquareWave& channel)
{
	int16_t sample(0);

	if (channel.currentVolume && channel.remainingLength > 0)
	{
		int pos = channel.squareStep / (channel.samplesPerDutyBit);

		sample = channel.currentVolume * channel.currentVolume;
		sample = sample * sample;
		if (!((channel.dutyMask >> (7 - pos)) & 1))
		{
			sample = -sample;
		}

		++channel.squareStep;
		if (channel.squareStep > (channel.samplesPerDutyBit * 8))
		{
			channel.squareStep = 0;
		}
	}
	return sample;
}

void
GameboySound::prepChannel(SquareWave& channel)
{
	uint16_t freqCode(channel.freqHigh & 0x7);
	freqCode <<= 8;
	freqCode |= channel.freqLow;

	double freq = 131072 / (2048.0 - freqCode);

	if (channel.sweep & 0x3)
	{
		if (channel.sweepDelay >= (channel.sweep & 0x3))
		{
			channel.sweepDelay = 0;
			++channel.sweepCount;
		}
		else
		{
			++channel.sweepDelay;
		}

		double freqMod = freq / (1 << channel.sweepCount);

		if (channel.sweep & 0x8)
		{
			freq -= freqMod;
		}
		else
		{
			freq += freqMod;
		}
	}

	if (channel.volume & 0x3)
	{
		// volumeDelay in 1/64 of a sec. x 2, as this callback occurs every
		// 1/128sec.
		if (channel.volumeDelay >= ((channel.volume & 0x3) * 2))
		{
			channel.volumeDelay = 0;
			if (channel.volume & 0x8)
			{
				if (channel.currentVolume < 0xf)
				{
					++channel.currentVolume;
				}
			}
			else
			{
				if (channel.currentVolume != 0)
				{
					--channel.currentVolume;
				}

			}
		}
		else
		{
			++channel.volumeDelay;
		}
	}

	if (channel.freqHigh & 0x40)
	{
		if (channel.remainingLength > 0) --channel.remainingLength;
	}

	channel.samplesPerDutyBit = (SAMPLE_RATE/freq) / 8;
}

int16_t
GameboySound::sampleChannel4()
{
	int16_t sample(0);

	if (m_channel4.currentVolume && m_channel4.remainingLength > 0)
	{
		sample = m_channel4.currentVolume * m_channel4.currentVolume;
		sample = sample * sample;
		if (!m_channel4.high)
		{
			sample = -sample;
		}

		++m_channel4.currentCounter;
		if (m_channel4.currentCounter > m_channel4.samplesPerStep)
		{
			m_channel4.high = rand() & 0x1;
		}
	}
	return sample;
}


void
GameboySound::prepChannel4()
{
	Noise& channel(m_channel4);

	if (channel.volume & 0x3)
	{
		// volumeDelay in 1/64 of a sec. x 2, as this callback occurs every
		// 1/128sec.
		if (channel.volumeDelay >= ((channel.volume & 0x3) * 2))
		{
			channel.volumeDelay = 0;
			if (channel.volume & 0x8)
			{
				if (channel.currentVolume < 0xf)
				{
					++channel.currentVolume;
				}
			}
			else
			{
				if (channel.currentVolume != 0)
				{
					--channel.currentVolume;
				}

			}
		}
		else
		{
			++channel.volumeDelay;
		}
	}

	if (channel.counter & 0x40)
	{
		if (channel.remainingLength > 0) --channel.remainingLength;
	}

	double ratio = channel.poly & 0x3;
	if (ratio == 0) ratio = 0.5;
	double freq = 524288 / ratio / std::pow(2.0, (channel.poly >> 4) + 1);
	channel.samplesPerStep = (SAMPLE_RATE/freq);
}


void
GameboySound::prepChannel3()
{
	uint16_t freqCode(m_channel3.freqHigh & 0x7);
	freqCode <<= 8;
	freqCode |= m_channel3.freqLow;

	double freq = 65536 / (2048.0 - freqCode);

	if (m_channel3.freqHigh & 0x40)
	{
		if (m_channel3.remainingLength > 0) --m_channel3.remainingLength;
	}

	m_channel3.samplesPerNibble = (SAMPLE_RATE/freq) / 32;
}

int16_t
GameboySound::sampleChannel3()
{
	int16_t sample(0);

	if 	(m_channel3.remainingLength > 0 &&
		(m_channel3.enabled & 0x80)
		)
	{
		int volume((m_channel3.volume >> 5) & 0x3);

		uint8_t nibble = m_channel3.waveTable[m_channel3.wavePos / 2];

		if (m_channel3.wavePos & 0x1)
		{
			nibble &= 0xf;
		}
		else
		{
			nibble >>= 4;
		}

		if (volume == 0)
		{
			nibble = 0;
		}
		else if (volume == 2)
		{
			nibble >>= 1;
		}
		else if (volume == 3)
		{
			nibble >>= 3;
		}

		// Ok, nibble is now between 0 and 0xf. Map this to full range.
		sample = (static_cast<int16_t>(nibble) - 8) * 4096;

		++m_channel3.waveSamples;
		if (m_channel3.waveSamples > m_channel3.samplesPerNibble * (m_channel3.wavePos + 1))
		{
			m_channel3.wavePos++;
			if (m_channel3.wavePos >= 32)
			{
				m_channel3.waveSamples = 0;
				m_channel3.wavePos = 0;
			}
		}

	}
	return sample;
}
