//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#ifndef GLBOY_GAMEBOYSOUND_HH
#define GLBOY_GAMEBOYSOUND_HH

#include "glBoy.hh"
#include "Core.hh"

#include "MemoryMap.hh"

#include <SDL/SDL.h>

namespace glBoy
{

class GameboySound : public MemoryMap::Memory
{
public:
	GameboySound(Core& core);

	void periodicUpdate(Core::CallbackId id);

	virtual uint8_t read8(uint16_t address);
	virtual void write8(uint16_t address, uint8_t value);

	void mixSounds(Uint8* stream, int len);
private:
	int16_t getSample(uint8_t volume);

	struct SquareWave
	{
		SquareWave() : sweep(0), duty(0), volume(0), freqLow(0), freqHigh(0)
		{
			reset();
		}

		void reset()
		{
			sweepDelay = 0;
			sweepCount = 0;
			volumeDelay = 0;
			currentVolume = volume >> 4;
			squareStep = 0;

			switch (duty >> 6)
			{
			case 0: dutyMask = 0x8; break;
			case 1: dutyMask = 0xC0; break;
			case 2: dutyMask = 0xF0; break;
			case 3: dutyMask = 0xFC; break;
			};

			// x 2, since callback @ 128Hz, not 256Hz.
			if (freqHigh & 0x40)
			{
				// Add 1, as we immediately decrement in prep method
				// without actually playing anything.
				remainingLength = ((64 - (duty & 0x3f)) / 2) + 1;
			}
			else
			{
				remainingLength = 255;
			}
		}
		// Regs.
		uint8_t sweep;
		uint8_t duty;
		uint8_t volume;
		uint8_t freqLow;
		uint8_t freqHigh;

		// Runtime
		int sweepDelay;
		uint32_t sweepCount;

		int volumeDelay;
		int currentVolume;

		uint8_t dutyMask;
		double samplesPerDutyBit;
		int squareStep;

		int remainingLength;
	};
	int16_t sample(SquareWave& channel);
	void prepChannel(SquareWave& channel);

	SquareWave m_channel1;
	SquareWave m_channel2;

	struct Noise
	{
		Noise() : length(0), volume(0), poly(0), counter(0) {reset(); }
		uint8_t length;
		uint8_t volume;
		uint8_t poly;
		uint8_t counter;

		int currentCounter;
		int samplesPerStep;
		int volumeDelay;
		int currentVolume;

		int remainingLength;

		bool high;

		void reset()
		{
			volumeDelay = 0;
			currentVolume = volume >> 4;

			currentCounter = 0;

			remainingLength = (64 - (length & 0x3f)) / 2;

			high = false;
		}
	};

	Noise m_channel4;
	int16_t sampleChannel4();
	void prepChannel4();

	struct Wave
	{
		Wave() : enabled(0), length(0), volume(0), freqLow(0), freqHigh(0)
		{
			reset();
		}

		void reset()
		{
			wavePos = 0;
			waveSamples = 0;

			// x 2, since callback @ 128Hz, not 256Hz.
			if (freqHigh & 0x40)
			{
				// Add 1, as we immediately decrement in prep method
				// without actually playing anything.
				remainingLength = ((256 - length) / 2) + 1;
			}
			else
			{
				remainingLength = 255;
			}
		}
		// Regs.
		uint8_t enabled;
		uint8_t length;
		uint8_t volume;
		uint8_t freqLow;
		uint8_t freqHigh;

		uint8_t waveTable[16];

		// Runtime
		int wavePos;
		int waveSamples;
		int remainingLength;
		float samplesPerNibble;
	};

	Wave m_channel3;
	int16_t sampleChannel3();
	void prepChannel3();

	uint8_t m_channelControl;
	uint8_t m_outputTerminal;
};

} // namespace glBoy

#endif
