//	Copyright (C) 2011 Michael McMaster <michael@codesrc.com>
//
//	This file is part of glBoy.
//
//	glBoy is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	glBoy is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with glBoy.  If not, see <http://www.gnu.org/licenses/>.

#include "glBoy.hh"
#include "GameboyTimer.hh"
#include "Log.hh"

#include <algorithm>
#include <cassert>

#include <time.h>

using namespace glBoy;

GameboyTimer::GameboyTimer(
	Core& core,
	const std::function<void(void)>& setFrameSkip,
	const std::function<void(void)>& clearFrameSkip
	) :
	m_core(core),
	m_setSkipFrame(setFrameSkip),
	m_clearSkipFrame(clearFrameSkip)
{
	memset(m_reg, 0, 4);

	using namespace std::placeholders;
	core.registerPeriodic(
		256,
		0,
		std::bind(&glBoy::GameboyTimer::tick, this, _1)
		);

	timespec res;
	clock_getres(CLOCK_REALTIME, &res);

	// Make sure we can perform high-resolution sleeps
// TODO resolve this
// meh. high resolution not all that necessary.
	m_clockSleepDelay = std::min(res.tv_nsec / 60000, 10l);
	//m_clockSleepDelay = res.tv_nsec / 60000;
	m_clockSleepCounter = m_clockSleepDelay;

/*
	if (m_clockSleepDelay)
	{
		Log::Instance()(Log::DBG_WARNING) <<
			"Timer delay set to " << m_clockSleepDelay <<
			" nanoseconds. Enable HR timers for improved emulation" << std::endl;
	}
*/

	clock_gettime(CLOCK_REALTIME, &m_lastSlept);
}

// Call this every 256 clock cycles. (16384Hz)

void
GameboyTimer::tick(Core::CallbackId)
{
	++m_reg[DIV];

	if (m_reg[TAC] & 0x4) // Timer enabled
	{
		int counter(m_reg[TIMA]);
		switch (m_reg[TAC] & 0x3)
		{
		case 0: // 4096Hz
			if (m_reg[DIV] % 4 == 0) ++counter;
			break;
		case 1: // 268400Hz
			counter += 16;
			break;
		case 2: // 65536Hz
			counter += 4;
			break;
		case 3: // 16384Hz
			++counter;
			break;
		}

		if (counter > 255)
		{
			m_reg[TIMA] = m_reg[TMA];
			m_core.raiseInterrupt(0x50);
		}
		else
		{
			m_reg[TIMA] = counter;
		}
	}

	m_lastSlept.tv_nsec += 61035;
	if (m_lastSlept.tv_nsec > 1e9)
	{
		m_lastSlept.tv_nsec -= 1e9;
		m_lastSlept.tv_sec++;
	}

	if (m_clockSleepCounter == 0)
	{
		timespec now;
		clock_gettime(CLOCK_REALTIME, &now);
		if (
			(m_lastSlept.tv_sec > now.tv_sec) ||
				(
					(m_lastSlept.tv_sec == now.tv_sec) &&
					(m_lastSlept.tv_nsec > now.tv_nsec)
				)
			)
		{
			m_clearSkipFrame();
			clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &m_lastSlept, NULL);
		}
		else
		{
			m_setSkipFrame();
		}
		m_clockSleepCounter = m_clockSleepDelay;
	}
	else
	{
		m_clockSleepCounter--;
	}

}

uint8_t
GameboyTimer::read8(uint16_t address)
{
	return m_reg[address];
}

void
GameboyTimer::write8(uint16_t address, uint8_t value)
{
	if (!address)
	{
		// Writing any value to DIV resets it to zero.
		m_reg[DIV] = 0;
	}
	else
	{
		m_reg[address] = value;
	}
}

